--------------------------------------------------------------------------------
-- House of Winters - Quest Journal
--------------------------------------------------------------------------------
-- Exposes the following to the global scope:
--     HoW_QJ               (addon namespace)
--     HoW_QJ_CT_*          (addon controls)
--     HoW_Settings         (addon settings page)
--     HoWUI_vars           (saved variables)
--
-- Provides the following slash commands:
--     /qj
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Setup module global variable.
--------------------------------------------------------------------------------

HoW_QJ = {
    name         = "HoW_QuestJournal",
    displayName  = "Quest Journal",
    prefix       = "HoW_QJ_CT_",
    author       = "Kentarii",
    version      = "1.2.0",
    link         = "http://www.esoui.com/",
    command      = "/qj",
}

--------------------------------------------------------------------------------
-- Libraries, defaults and constants.
--------------------------------------------------------------------------------

local lam = LibStub:GetLibrary("LibAddonMenu-1.0")
local url = LibStub:GetLibrary("LibURLEncode-1.0")

--
-- Module defaults.
--
local defaults = {
    visible = true,
    faded = false,
    alpha = 65,
    anchor = {
        point = BOTTOMRIGHT,
        relativePoint = BOTTOMRIGHT,
        offsetX = -10,
        offsetY = -10,
    },
    colors = {
        header = "EECA2A",
        section = "FF9933",
        text = "FFE3A1",
        value = "F0F0F0",
        abandoned = "C00000",
        active = "6699FF",
        complete = "208020",
        unknown = "E0E0E0",
    },
    dimension = {
        width = 400,
        height = 300,
    },
    mapPinSize = 32,
    mapPinLayer = 105,
    mapPinTextures = "default_64px",
    filters = {
        pve = {
            abandoned = true,
            active    = true,
            complete  = true,
            unknown   = false,
        },
        pvp = {
            abandoned = true,
            active    = true,
            complete  = true,
            unknown   = false,
        },
    },
    quests = {},
}

--
-- Quest states.
--
local STATE = {
    unknown = 0,
    active = 1,
    complete = 2,
    abandoned = 3,
}

--
-- Supported fonts
--
local FONTS = {
   "ZoFontBookLetter",
   "ZoFontBookNote",
   "ZoFontBookPaper",
   "ZoFontBookRubbing",
   -- "ZoFontBookScroll",
   "ZoFontBookSkin",
   -- "ZoFontBookTablet",
   "ZoFontChat",
   "ZoFontGame",
   "ZoFontGameBold",
   "ZoFontGameLarge",
   "ZoFontGameLargeBold",
   "ZoFontGameLargeBoldShadow",
   "ZoFontGameMedium",
   "ZoFontGameOutline",
   "ZoFontGameShadow",
   "ZoFontGameSmall",
   "ZoFontWinT1",
   "ZoFontWinT2",
}

--------------------------------------------------------------------------------
-- Addon init and event/command handling functions.
--------------------------------------------------------------------------------


--- Event handler for EVENT_ADD_ON_LOADED.
-- @return void
function HoW_QJ.AddOnLoaded(eventCode, addOnName)
    if addOnName ~= HoW_QJ.name then return end

    -- Register the slash command handler.
    SLASH_COMMANDS[HoW_QJ.command] = HoW_QJ.CommandHandler

    -- Initialize addon.
    HoW_QJ:Init()

    -- Add addon settings page.
    HoW_QJ.Settings()
end

--- Initialize HoW_QJ.
-- @return void
function HoW_QJ:Init()
    -- Load saved vars.
    HoW_QJ.vars = ZO_SavedVars:New("HoWUI_vars", 1, "QuestJournal", defaults)
    -- Add settings which has come after version 1 was introduced.
    if not HoW_QJ.vars.filters then
        HoW_QJ.vars.filters = defaults.filters
    end

    if not HoW_QJ.vars.currentZoneName then
        HoW_QJ.vars.currentZoneName = GetUnitZone("player")
    end
    HoW_QJ.tmp = {}

    -- Update control from vars.
    HoW_QJ_CT:ClearAnchors()
    HoW_QJ:SetAnchor(HoW_QJ.vars.anchor)
    HoW_QJ.SetAlpha(HoW_QJ.vars.alpha)
    HoW_QJ:SetDimensions(HoW_QJ.vars.dimension)
    HoW_QJ:SetFaded(HoW_QJ.vars.faded)

    -- Load first time data.
    HoW_QJ.Update()

    if HoW_QJ.vars.visible then
        HoW_QJ:OpenWindow()
    end

    -- Initialize mappins
    for pinKey, pinType in pairs(HoW_QJ.PINS.names) do
        ZO_WorldMap_AddCustomPin(pinType, HoW_QJ.PINS.func.createPins[pinKey], nil, {
            level = HoW_QJ.GetMapPinLayer(),
            texture = HoW_QJ.PINS.textures[HoW_QJ.vars.mapPinTextures][pinKey],
            size = HoW_QJ.GetMapPinSize(),
        }, HoW_QJ.PINS.func.tooltip[pinKey])
        ZO_WorldMap_SetCustomPinEnabled(_G[pinType], true)
        ZO_WorldMap_RefreshCustomPinsOfType(_G[pinType])
        -- Add map filters, unknown is currently not supported since no data exists.
        if not (pinKey == "unknown") then
            HoW_QJ.PINS.func.AddWorldMapFilters(WORLD_MAP_FILTERS.pvePanel, "pve", pinKey, pinType, HoW_QJ.PINS.filterText[pinKey], HoW_QJ.vars.filters.pve[pinKey])
            HoW_QJ.PINS.func.AddWorldMapFilters(WORLD_MAP_FILTERS.pvpPanel, "pvp", pinKey, pinType, HoW_QJ.PINS.filterText[pinKey], HoW_QJ.vars.filters.pvp[pinKey])
        end
    end

    -- Bind event handlers.
    EVENT_MANAGER:RegisterForEvent(HoW_QJ.name, EVENT_QUEST_ADDED,              HoW_QJ.OnEventQuestAdded)
    EVENT_MANAGER:RegisterForEvent(HoW_QJ.name, EVENT_QUEST_REMOVED,            HoW_QJ.OnEventQuestRemoved)
    EVENT_MANAGER:RegisterForEvent(HoW_QJ.name, EVENT_QUEST_COMPLETE_DIALOG,	HoW_QJ.OnEventQuestCompleteDialog)
    EVENT_MANAGER:RegisterForEvent(HoW_QJ.name, EVENT_QUEST_COMPLETE,  			HoW_QJ.OnEventQuestComplete)

    ZO_PreHookHandler(ZO_GameMenu_InGame, "OnShow", function()
        HoW_QJ:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_GameMenu_InGame, "OnHide", function()
        HoW_QJ:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_InteractWindow, "OnShow", function()
        HoW_QJ:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_InteractWindow, "OnHide", function()
        HoW_QJ:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_KeybindStripControl, "OnShow", function()
        HoW_QJ:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_KeybindStripControl, "OnHide", function()
        HoW_QJ:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_MainMenuCategoryBar, "OnShow", function()
        HoW_QJ:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_MainMenuCategoryBar, "OnHide", function()
        HoW_QJ:ShowIfVisible()
    end)
end

--- Event handler for EVENT_QUEST_ADDED.
-- @param eventCode int
-- @param journalIndex int
-- @param questName string
-- @param objectiveName string
-- @return void
function HoW_QJ.OnEventQuestAdded(eventCode, journalIndex, questName, objectiveName)
    -- Quest was started by user, update states.
    local _, backgroundText, _, _, _, _, _, questLevel, _, questType = GetJournalQuestInfo(journalIndex)
    local zoneName, _, zoneIndex = GetJournalQuestLocationInfo(journalIndex)
    local posX, posY = GetMapPlayerPosition("player")
    local meta = {
        quest = {
            name = questName,
            description = backgroundText,
            level = questLevel,
            type = questType,
        },
        zone = {
            index = zoneIndex,
            name = zoneName,
        },
        agent = {
            name = GetNameOfGameCameraQuestToolTarget(),
            zone = GetUnitZone("player"),
            mapname = GetMapName(),
            pin = {
                x = posX,
                y = posY,
            },
        },
    }
    HoW_QJ:UpdateQuest(questName, zoneName, STATE.active, meta)
    d(string.format("Quest started: |c%s%s|r", HoW_QJ.vars.colors.active, questName))
    if HoW_QJ.DB.quests[zoneName] and HoW_QJ.vars.visible then
        HoW_QJ.vars.currentZoneName = zoneName
        -- HoW_QJ.vars.currentQuestName = questName
        HoW_QJ:UpdateWindow(true)
    end

    -- Refresh affected map pins
    ZO_WorldMap_RefreshCustomPinsOfType(_G[HoW_QJ.PINS.names.abandoned])
    ZO_WorldMap_RefreshCustomPinsOfType(_G[HoW_QJ.PINS.names.active])
end

--- Event handler for EVENT_QUEST_COMPLETE_DIALOG.
-- @param eventCode int
-- @param journalIndex int
-- @return void
function HoW_QJ.OnEventQuestCompleteDialog(eventCode, journalIndex)
    -- Use this to get zoneIndex and zoneName for quest complete event before it removed from quest journal.
    local questName, _, _, _, _, _, _, questLevel, _, questType = GetJournalQuestInfo(journalIndex)
    local _, _, _, _, backgroundText, _ = GetJournalQuestEnding(journalIndex)
    local zoneName, _, zoneIndex = GetJournalQuestLocationInfo(journalIndex)
    local posX, posY = GetMapPlayerPosition("player")
    local meta = {
        quest = {
            name = questName,
            description = backgroundText,
            level = questLevel,
            type = questType,
        },
        zone = {
            index = zoneIndex,
            name = zoneName,
        },
        agent = {
            name = GetNameOfGameCameraQuestToolTarget(),
            zone = GetUnitZone("player"),
            mapname = GetMapName(),
            pin = {
                x = posX,
                y = posY,
            },
        },
    }
    -- d(meta)
    HoW_QJ.tmp[questName] = meta
end

--- Event handler for EVENT_QUEST_COMPLETE.
-- @param eventCode int
-- @param questName string
-- @param xpGained int
-- @return void
function HoW_QJ.OnEventQuestComplete(eventCode, questName, xpGained)
    -- Quest was completed by user, update states.
    -- Use cached data from OnEventQuestCompleteDialog
    local meta = HoW_QJ.tmp[questName]
    local zoneName = meta.zone.name
    HoW_QJ:UpdateQuest(questName, zoneName, STATE.complete, meta)
    d(string.format("Quest completed: |c%s%s|r", HoW_QJ.vars.colors.complete, questName))
    HoW_QJ.tmp[questName] = nil
    if HoW_QJ.DB.quests[zoneName] and HoW_QJ.vars.visible then
        HoW_QJ.vars.currentZoneName = zoneName
        -- HoW_QJ.vars.currentQuestName = questName
        HoW_QJ:UpdateWindow(true)
    end

    -- Refresh affected map pins
    ZO_WorldMap_RefreshCustomPinsOfType(_G[HoW_QJ.PINS.names.active])
    ZO_WorldMap_RefreshCustomPinsOfType(_G[HoW_QJ.PINS.names.complete])
end

--- Event handler for EVENT_QUEST_REMOVED.
-- @param eventCode int
-- @param isCompleted bool
-- @param journalIndex int
-- @param questName string
-- @param zoneIndex int
-- @param poiIndex int
-- @return void
function HoW_QJ.OnEventQuestRemoved(eventCode, isCompleted, journalIndex, questName, zoneIndex, poiIndex)
    if not isCompleted then
        -- Quest was abandoned by user, update states.
        local zoneName = HoW_QJ.DB.zones[""..zoneIndex]

        if zoneName then
            HoW_QJ:UpdateQuest(questName, zoneName, STATE.abandoned)
            d(string.format("Quest abandoned: |c%s%s|r", HoW_QJ.vars.colors.abandoned, questName))
            if HoW_QJ.vars.visible then
                HoW_QJ.vars.currentZoneName = zoneName
                HoW_QJ:UpdateWindow(true)
            end
        else
            -- TODO check all quest types?
        end


        -- Refresh affected map pins
        ZO_WorldMap_RefreshCustomPinsOfType(_G[HoW_QJ.PINS.names.abandoned])
        ZO_WorldMap_RefreshCustomPinsOfType(_G[HoW_QJ.PINS.names.active])
    end
end

--- Event handler for OnMoveStop to reposition window per character.
-- @return void
function HoW_QJ:OnMoveStop()
    local isValidAnchor, point, relativeTo, relativePoint, offsetX, offsetY = HoW_QJ_CT:GetAnchor()
    if (isValidAnchor) then
        HoW_QJ.vars.anchor.point = point
        HoW_QJ.vars.anchor.relativePoint = relativePoint
        HoW_QJ.vars.anchor.offsetX = offsetX
        HoW_QJ.vars.anchor.offsetY = offsetY
    end
end

--- Event handler for OnResizeStop to resize window per character.
-- @return void
function HoW_QJ:OnResizeStop()
    local width, height = HoW_QJ_CT:GetDimensions()
    if width and height then
        HoW_QJ.vars.dimension.width = width
        HoW_QJ.vars.dimension.height = height
    end
    -- Resize might cause anchor change as well.
    HoW_QJ:OnMoveStop()
end

--- Event handler for OnMouseWheel to scroll TextBuffer control.
-- @param delta int scroll up/down step value.
-- @return void
function HoW_QJ:OnMouseWheel(delta)
    local scrollPosition = HoW_QJ_CT_TextBuffer:GetScrollPosition()
    local visible = HoW_QJ:CalcVisibleLines()
    local lines = HoW_QJ_CT_TextBuffer:GetNumHistoryLines()
    HoW_QJ_CT_TextBuffer:MoveScrollPosition(delta)
    -- TODO doesn't work that well with wrapped lines..
    --if delta < 0 then
    --    HoW_QJ_CT_TextBuffer:MoveScrollPosition(delta)
    --elseif delta > 0 then
    --    if (lines - visible) > scrollPosition then
    --        HoW_QJ_CT_TextBuffer:MoveScrollPosition(delta)
    --    else
    --        HoW_QJ:ScrollToFirstLine()
    --    end
    --end
end

--- Event handler for OnLinkClicked in TextBuffer control.
-- @param self
-- @param linkData string the link url:     url:SomeLinkTargetOrUrl[SomeLinkText]
-- @param linkText string the link text:    [SomeLinkText]
-- @param button int the mousebutton clicked, 1 = left, 2 = right, 3 = middle
-- @param ctrl bool true if ctrl was pressed during click.
-- @param alt bool true if alt was pressed during click.
-- @param shift bool true if shift was pressed during click.
-- @param command bool true if command (Mac I assume) was pressed during click.
-- @return void
function HoW_QJ:OnLinkClicked(_self, linkData, linkText, button, ctrl, alt, shift, command)
    local linkStyle, linkType, arg1, arg2 = zo_strsplit(":", linkData)
    if linkType == "quest" then
        local zoneName = arg1
        local questName = url:decode(arg2)
        -- d(string.format("zoneName: %s, questName: %s", zoneName, questName))
        if HoW_QJ.DB.quests[zoneName] and HoW_QJ.DB.quests[zoneName][questName] then
            if button == 1 then
                HoW_QJ.vars.currentQuestName = questName
                HoW_QJ:UpdateWindow(true)
            elseif button == 2 then
                if HoW_QJ.vars.quests[zoneName] and HoW_QJ.vars.quests[zoneName][questName] and HoW_QJ.vars.quests[zoneName][questName].state == STATE.complete then
                    HoW_QJ:UpdateQuest(questName, zoneName, STATE.unknown)
                else
                    HoW_QJ:UpdateQuest(questName, zoneName, STATE.complete)
                end
            end
            local scrollPosition = HoW_QJ_CT_TextBuffer:GetScrollPosition()
            HoW_QJ:UpdateWindow(true)
            HoW_QJ_CT_TextBuffer:SetScrollPosition(scrollPosition)
        end
    elseif linkType == "filter" then
        local state = tonumber(arg1)
        HoW_QJ.vars.currentQuestName = nil
        -- d(string.format("state: %d", state))
        if button == 1 then
            if state == STATE.unknown or state == STATE.active or state == STATE.complete then
                HoW_QJ.vars.currentFilter = state
            else
                HoW_QJ.vars.currentFilter = nil
            end
            HoW_QJ:UpdateWindow(true)
        end
    elseif linkType == "zone" then
        local zoneName = arg1
        HoW_QJ.vars.currentQuestName = nil
        if button == 1 then
            if zoneName == "select" then
                HoW_QJ.vars.currentZoneName = ""
                HoW_QJ_CT_TextLabel:SetText(string.format(
                    "|c%sPlease select a zone:",
                    HoW_QJ.vars.colors.text
                ))
            else
                HoW_QJ.vars.currentZoneName = zoneName
            end
            HoW_QJ:UpdateWindow(true)
        end
    end
end

function HoW_QJ:CalcVisibleLines()
    local width, height = HoW_QJ_CT_TextBuffer:GetDimensions()
    return math.floor(height / 24)
end

function HoW_QJ:ScrollToFirstLine()
    local visible = HoW_QJ:CalcVisibleLines()
    local lines = HoW_QJ_CT_TextBuffer:GetNumHistoryLines()
    HoW_QJ_CT_TextBuffer:SetScrollPosition(lines - visible)
end

function HoW_QJ:ScrollToLine(line)
    local visible = HoW_QJ:CalcVisibleLines()
    local lines = HoW_QJ_CT_TextBuffer:GetNumHistoryLines()

    HoW_QJ_CT_TextBuffer:SetScrollPosition(lines - (visible + line - 1))
end

--- Addon command handler.
-- @param text the command line argument (original case)
-- @return void
function HoW_QJ.CommandHandler(text)
    local cmd = string.lower(text)

    if cmd == "show" then
        HoW_QJ.Update()
        HoW_QJ.OpenWindow({["zoneName"] = GetUnitZone("player")})
    elseif cmd == "hide" then
        HoW_QJ:CloseWindow()
    elseif cmd == "import" then
        HoW_QJ:ImportEsohead(false)
    elseif cmd == "import-esohead" then
        HoW_QJ:ImportEsohead(true)
    else
        d("----------------------------------------")
        d("|cEECA2A"..HoW_QJ.displayName.."|r |c6699FF"..HoW_QJ.version.."|r by |cFF9933"..HoW_QJ.author.."|r")
        d("----------------------------------------")
        d("Commands:")
        d(HoW_QJ.command.." show")
        d(HoW_QJ.command.." hide")
        d("----------------------------------------")
    end
end

--------------------------------------------------------------------------------
-- Import data from external sources.
--------------------------------------------------------------------------------

--- Try to import quests from Esohead.lua
-- @param doImport boolean whether or not to actually do the import
-- @return void
function HoW_QJ:ImportEsohead(doImport)
    local esoheadDefaults = {data = {}}
    local esoheadQuests = ZO_SavedVars:NewAccountWide("Esohead_SavedVariables", 2, "quest", esoheadDefaults)
    local questNames = {}
    local found = 0

    -- Map Esohead quests to a simple questName => zoneName table
    -- ZoneName in Esohead does not always match the zonename in QJ.
    if esoheadQuests.data then
        for zoneName, quests in orderedPairs(esoheadQuests.data) do
            for questName, markers in orderedPairs(quests) do
                questNames[questName] = zoneName
                found = found + 1
            end
        end
    end

    HoW_QJ_CT:SetHidden(false)
    HoW_QJ:PrintHeader("Esohead Import Preview")
    HoW_QJ_CT_TextLabel:SetText(string.format("This is only a preview,\nto import data, please run:\n/qj import-esohead"))
    HoW_QJ_CT_TextBuffer:Clear()
    if found > 100 then
        HoW_QJ_CT_TextBuffer:SetMaxHistoryLines(found + 10)
    end

    local matches = 0
    local importCount = 0
    for zoneName, quests in orderedPairs(HoW_QJ.DB.quests) do
        for questName, questData in orderedPairs(quests) do
            -- Found matching questName
            if questNames[questName] then
                matches = matches + 1
                local zoneMatch = (zoneName == questNames[questName])
                local questColor = HoW_QJ.vars.colors.unknown
                local questState = STATE.unknown

                -- Check if quest is already tracked by Quest Journal
                if HoW_QJ.vars.quests[zoneName] and HoW_QJ.vars.quests[zoneName][questName] then
                    local quest = HoW_QJ.vars.quests[zoneName][questName]
                    questState = quest.state
                    if quest.state == STATE.active then
                        questColor = HoW_QJ.vars.colors.active
                    elseif quest.state == STATE.complete then
                        questColor = HoW_QJ.vars.colors.complete
                    elseif quest.state == STATE.abandoned then
                        questColor = HoW_QJ.vars.colors.abandoned
                    end
                end

                HoW_QJ_CT_TextBuffer:AddMessage(string.format(
                    "|c%s%s|r L%d (%s: %s)",
                    questColor,
                    questName,
                    questData.level,
                    zoneName,
                    tostring(zoneMatch)
                ), 1, 1, 1, 1);

                if questState == STATE.unknown then
                    importCount = importCount + 1
                end
                if doImport and questState == STATE.unknown then
                    HoW_QJ:UpdateQuest(questName, zoneName, STATE.complete)
                end
            end
        end
    end
    if doImport then
        HoW_QJ_CT_TextBuffer:AddMessage(string.format("Imported %d quests.", importCount))
    else
        HoW_QJ_CT_TextBuffer:AddMessage(string.format("Found %d quests in Esohead.lua, matched %d of those with QJ quests, %d new quests will be imported.", found, matches, importCount))
    end
end

--------------------------------------------------------------------------------
-- Setup module control panel section.
--------------------------------------------------------------------------------

--- Addon settings.
-- @return void
function HoW_QJ.Settings()
    local controlPanel = HoW_Settings

    -- Create HoW settings page.
    if not controlPanel then
        controlPanel = lam:CreateControlPanel("HoW_Settings", "|cEECA2AHoW Addons|r")
    end

    -- Add module settings.
    lam:AddHeader(controlPanel, HoW_QJ.prefix.."SettingsHeader", HoW_QJ.displayName)
    lam:AddCheckbox(controlPanel, HoW_QJ.prefix.."SettingsVisible", "Visible", "Toggle Quest Journal visibility", HoW_QJ.IsVisible, HoW_QJ.SetVisible)
    lam:AddDropdown(controlPanel, HoW_QJ.prefix.."SettingsFont", "Font", "Change font family", FONTS, HoW_QJ.GetFont, HoW_QJ.SetFont)
    lam:AddSlider(controlPanel, HoW_QJ.prefix.."SettingsScale", "Scale", "Change window scaling", 75, 100, 1, HoW_QJ.GetScale, HoW_QJ.SetScale)
    lam:AddSlider(controlPanel, HoW_QJ.prefix.."SettingsAlpha", "Alpha", "Change background transparency", 0, 100, 1, HoW_QJ.GetAlpha, HoW_QJ.SetAlpha)

    lam:AddSlider(controlPanel, HoW_QJ.prefix.."SettingsMapPinSize", "Map pin size", "Change quest map pin size", 16, 64, 1, HoW_QJ.GetMapPinSize, HoW_QJ.SetMapPinSize)
    lam:AddSlider(controlPanel, HoW_QJ.prefix.."SettingsMapPinLayer", "Map pin layer", "Change quest map pin layer", 10, 200, 5, HoW_QJ.GetMapPinLayer, HoW_QJ.SetMapPinLayer)

    lam:AddDescription(controlPanel, HoW_QJ.prefix.."AboutDescription", "Chat command: |c66FF99"..HoW_QJ.command.."|r\n\n|cEECA2A"..HoW_QJ.displayName.."|r |c6699FF"..HoW_QJ.version.."|r by |cFF9933"..HoW_QJ.author.."|r")
end

function HoW_QJ.IsVisible()
    return HoW_QJ.vars and HoW_QJ.vars.visible
end

function HoW_QJ.SetVisible(checkbox)
    HoW_QJ.vars.visible = checkbox
end

function HoW_QJ.GetFont()
    if HoW_QJ.vars and HoW_QJ.vars.font then
        return HoW_QJ.vars.font
    else
        return "ZoFontGame"
    end
end

function HoW_QJ.SetFont(font)
    HoW_QJ.vars.font = font
    HoW_QJ_CT_HeaderLabel:SetFont(font)
    HoW_QJ_CT_TextLabel:SetFont(font)
    HoW_QJ_CT_TextBuffer:SetFont(font)
end

function HoW_QJ.GetScale()
    if HoW_QJ.vars and HoW_QJ.vars.scale then
        return HoW_QJ.vars.scale
    else
        return 100
    end
end

function HoW_QJ.SetScale(scale)
    HoW_QJ.vars.scale = scale
    HoW_QJ_CT:SetScale(scale / 100)
end

function HoW_QJ.GetAlpha()
    if HoW_QJ.vars and HoW_QJ.vars.alpha then
        return HoW_QJ.vars.alpha
    else
        return 65
    end
end

function HoW_QJ.SetAlpha(alpha)
    HoW_QJ.vars.alpha = alpha
    HoW_QJ_CT_Backdrop:SetAlpha(alpha / 100)
end

function HoW_QJ.GetMapPinSize()
    if HoW_QJ.vars and HoW_QJ.vars.mapPinSize then
        return HoW_QJ.vars.mapPinSize
    else
        return 32
    end
end

function HoW_QJ.SetMapPinSize(size)
    HoW_QJ.vars.mapPinSize = size

    for pinKey, pinType in pairs(HoW_QJ.PINS.names) do
        ZO_MapPin.PIN_DATA[_G[pinType]].size = size
        ZO_WorldMap_RefreshCustomPinsOfType(_G[pinType])
    end
end

function HoW_QJ.GetMapPinLayer()
    if HoW_QJ.vars and HoW_QJ.vars.mapPinLayer then
        return HoW_QJ.vars.mapPinLayer
    else
        return 110
    end
end

function HoW_QJ.SetMapPinLayer(layer)
    HoW_QJ.vars.mapPinLayer = layer

    for pinKey, pinType in pairs(HoW_QJ.PINS.names) do
        ZO_MapPin.PIN_DATA[_G[pinType]].level = layer
        ZO_WorldMap_RefreshCustomPinsOfType(_G[pinType])
    end
end

--------------------------------------------------------------------------------
-- Update functions.
--------------------------------------------------------------------------------

--- Read questlog and update entries.
-- @return void
function HoW_QJ.Update()
    local quests = GetNumJournalQuests()
    HoW_QJ.tmp = {}
    for i = 1, quests, 1 do
        local questName, backgroundText, _, _, _, _, _, questLevel, _, questType = GetJournalQuestInfo(i)
        local zoneName, _, zoneIndex = GetJournalQuestLocationInfo(i)
        local meta = {
            quest = {
                name = questName,
                description = backgroundText,
                level = questLevel,
                type = questType,
            },
            zone = {
                index = zoneIndex,
                name = zoneName,
            },
        }
        HoW_QJ:UpdateQuest(questName, zoneName, STATE.active, meta)
    end
end

--- Add/update quest entry in character quest journal history.
-- @param questName string
-- @param zoneName string
-- @param state int
-- @param meta table optional meta data sent with added and completed events.
-- @return void
function HoW_QJ:UpdateQuest(questName, zoneName, state, meta)
    if not questName then return end

    -- zoneName is not set, try using quest type instead.
    if not zoneName or zoneName == "" then
        if meta and meta.quest and meta.quest.type then
            zoneName = HoW_QJ.DB.types[tostring(meta.quest.type)]
            -- d("zonename set from type")
        else
            -- d("zonename could not be set")
            return
        end
    end

    -- Storing quests for all zones, even though they are not supported.
    if not HoW_QJ.vars.quests[zoneName] then HoW_QJ.vars.quests[zoneName] = {} end
    if not HoW_QJ.vars.quests[zoneName][questName] then HoW_QJ.vars.quests[zoneName][questName] = {} end
    HoW_QJ.vars.quests[zoneName][questName].state = state
    HoW_QJ.vars.quests[zoneName][questName].timestamp = GetTimeStamp()

    if state == STATE.active and meta then
        -- Can be updated from event quest added or from quest journal.
        -- Quest journal will not have the agent info.
        if not HoW_QJ.vars.quests[zoneName][questName].start then HoW_QJ.vars.quests[zoneName][questName].start = {} end
        if not HoW_QJ.vars.quests[zoneName][questName].start.quest and meta.quest then
            HoW_QJ.vars.quests[zoneName][questName].start.quest = meta.quest
            HoW_QJ.vars.quests[zoneName][questName].level = meta.quest.level
        end
        if meta.zone then HoW_QJ.vars.quests[zoneName][questName].start.zone = meta.zone end
        if meta.agent then HoW_QJ.vars.quests[zoneName][questName].start.agent = meta.agent end
        if not HoW_QJ.vars.quests[zoneName][questName].start.timestamp then HoW_QJ.vars.quests[zoneName][questName].start.timestamp = GetTimeStamp() end
    elseif state == STATE.complete and meta then
        -- Can only be updated from event complete experience.
        meta.timestamp = GetTimeStamp()
        HoW_QJ.vars.quests[zoneName][questName].finish = meta
        if not HoW_QJ.vars.quests[zoneName][questName].level then
            HoW_QJ.vars.quests[zoneName][questName].level = meta.quest.level
        end
    end
end

--- Get active/complete quests for selected zone for character
-- @param zoneName string the selected zone.
-- @return active, complete values as int.
function HoW_QJ:GetQuestStats(zoneName)
    local active = 0
    local complete = 0
    if HoW_QJ.vars.quests[zoneName] then
        for _, quest in pairs(HoW_QJ.vars.quests[zoneName]) do
            if quest.state == STATE.active then
                active = active + 1
            elseif quest.state == STATE.complete then
                complete = complete + 1
            end
        end
    end
    return active, complete
end

--- Get quest info.
-- @param zoneName string the selected zone.
-- @param questName string the selected quest.
-- @return quest map with quest info.
function HoW_QJ:GetQuestInfo(zoneName, questName)
    local quest = nil
    if HoW_QJ.vars.quests[zoneName] and HoW_QJ.vars.quests[zoneName][questName] then
        quest = HoW_QJ.vars.quests[zoneName][questName]
        if not quest.level and HoW_QJ.DB.quests[zoneName] and HoW_QJ.DB.quests[zoneName][questName] then
            quest.level = HoW_QJ.DB.quests[zoneName][questName].level
        end
    elseif HoW_QJ.DB.quests[zoneName] and HoW_QJ.DB.quests[zoneName][questName] then
        quest = HoW_QJ.DB.quests[zoneName][questName]
    end
    return quest
end

--------------------------------------------------------------------------------
-- Show quest journal window.
--------------------------------------------------------------------------------

--- Show QuestJournal window with all timers or timers for a type and/or character.
-- @param options table with options to show, types/character.
-- @return void
function HoW_QJ.OpenWindow(options)
    -- Remember window state.
    HoW_QJ.vars.visible = true

    -- Save module so UpdateWindow can access it.
    if (options and options.zoneName and HoW_QJ.DB.quests[options.zoneName]) then
        HoW_QJ.vars.currentZoneName = options.zoneName
    end

    -- Set font and scale on components
    local font = HoW_QJ.GetFont()
    HoW_QJ_CT_HeaderLabel:SetFont(font)
    HoW_QJ_CT_TextLabel:SetFont(font)
    HoW_QJ_CT_TextBuffer:SetFont(font)

    HoW_QJ:UpdateWindow(true)
    HoW_QJ_CT:SetHidden(false)
end

--- Update all timers or timers for a type and/or character.
--- Called from xml without args and OpenWindow.
-- @param force trigger update without buffering when opening up window to show content directly.
-- @return void
function HoW_QJ:UpdateWindow(force)
    if not force and not BufferReached("HoW_QJ.OnUpdate", 1) then return end

    -- Header
    HoW_QJ:PrintHeader(HoW_QJ.vars.currentZoneName)

    -- Zone/quest statistics
    local stats = HoW_QJ.DB.stats[HoW_QJ.vars.currentZoneName]
    if stats then
        HoW_QJ:PrintQuestStats(stats, HoW_QJ.vars.currentZoneName)
        if HoW_QJ.vars.currentQuestName then
            HoW_QJ:PrintQuestInfo(HoW_QJ.vars.currentZoneName, HoW_QJ.vars.currentQuestName)
        else
            HoW_QJ:PrintQuestList(HoW_QJ.vars.currentZoneName, HoW_QJ.vars.currentFilter)
        end
    else
        if not (HoW_QJ.vars.currentZoneName == "") then
            HoW_QJ_CT_TextLabel:SetText(string.format(
                "|c%sThis location is not supported,\nplease select another zone:",
                HoW_QJ.vars.colors.text
            ))
        end
        HoW_QJ:PrintZoneList()
    end
end

--- Print heading with addon name and optionally zonename.
-- @param zoneName string the current selected zoneName.
-- @return void
function HoW_QJ:PrintHeader(zoneName)
    if zoneName and not (zoneName == "") then
        HoW_QJ_CT_HeaderLabel:SetText(string.format(
            "|c%s%s\n|c%s|H%d:zone:select:|h%s|h",
            HoW_QJ.vars.colors.header,
            HoW_QJ.displayName,
            HoW_QJ.vars.colors.section,
            LINK_STYLE_DEFAULT,
            zoneName
        ))
    else
        HoW_QJ_CT_HeaderLabel:SetText(string.format(
            "|c%s%s",
            HoW_QJ.vars.colors.header,
            HoW_QJ.displayName
        ))
    end
end

--- Update TextLabel with quests stats.
-- @param stats map with total and required for achievement
-- @param zoneName string the current selected zoneName.
-- @return void
function HoW_QJ:PrintQuestStats(stats, zoneName)
    local active, complete = HoW_QJ:GetQuestStats(zoneName)
    local unknown = stats.total - (active + complete)

    HoW_QJ_CT_TextLabel:SetText(string.format(
        "|c%sTotal quests: |c%s|H%d:filter:-1:|h%d|h|c%s, achievement: |c%s%d|c%s\nAvailable: |c%s|H%d:filter:%d:|h%d|h|c%s, active: |c%s|H%d:filter:%d:|h%d|h|c%s, completed: |c%s|H%d:filter:%d:|h%d|h",
        HoW_QJ.vars.colors.text,
        HoW_QJ.vars.colors.value,
        LINK_STYLE_DEFAULT,
        stats.total,
        HoW_QJ.vars.colors.text,
        HoW_QJ.vars.colors.value,
        stats.required,
        HoW_QJ.vars.colors.text,
        HoW_QJ.vars.colors.unknown,
        LINK_STYLE_DEFAULT,
        STATE.unknown,
        unknown,
        HoW_QJ.vars.colors.text,
        HoW_QJ.vars.colors.active,
        LINK_STYLE_DEFAULT,
        STATE.active,
        active,
        HoW_QJ.vars.colors.text,
        HoW_QJ.vars.colors.complete,
        LINK_STYLE_DEFAULT,
        STATE.complete,
        complete
    ))
end

--- Update TextBuffer with quests.
-- @param zoneName string the current selected zoneName.
-- @param filter int optional filter to apply to list.
-- @return void
function HoW_QJ:PrintQuestList(zoneName, filter)
    local quests = HoW_QJ.DB.quests[zoneName]

    HoW_QJ_CT_TextBuffer:Clear()
    if quests then
        local i = 1
        for questName, q in orderedPairs(quests) do
            local questColor = HoW_QJ.vars.colors.unknown
            local questState = STATE.unknown
            if HoW_QJ.vars.quests[zoneName] and HoW_QJ.vars.quests[zoneName][questName] then
                local quest = HoW_QJ.vars.quests[zoneName][questName]
                questState = quest.state
                if quest.state == STATE.active then
                    questColor = HoW_QJ.vars.colors.active
                elseif quest.state == STATE.complete then
                    questColor = HoW_QJ.vars.colors.complete
                elseif quest.state == STATE.abandoned then
                    questColor = HoW_QJ.vars.colors.abandoned
                end
            end
            if not filter or filter == questState or (questState == STATE.abandoned and filter == STATE.unknown) then
                HoW_QJ_CT_TextBuffer:AddMessage(string.format(
                    "|c%s#%02d |c%s|H%d:quest:%s:%s:|h%s|h |c%s(lvl %d)",
                    HoW_QJ.vars.colors.text,
                    i,
                    questColor,
                    LINK_STYLE_DEFAULT,
                    zoneName, url:encode(questName),
                    questName,
                    HoW_QJ.vars.colors.text,
                    q.level
                ), 1, 1, 1, 1);
                i = i + 1
            end
        end
        HoW_QJ:ScrollToFirstLine()
    end
end

--- Update TextBuffer with zones.
-- @return void
function HoW_QJ:PrintZoneList()
    local zones = HoW_QJ.DB.stats

    HoW_QJ_CT_TextBuffer:Clear()
    local i = 1
    for zoneName, stats in orderedPairs(zones) do
        local _, complete = HoW_QJ:GetQuestStats(zoneName)
        HoW_QJ_CT_TextBuffer:AddMessage(string.format(
            "|c%s#%02d |c%s|H%d:zone:%s:|h%s|h |c%s(|c%s%d|c%s/|c%s%d|c%s)",
            HoW_QJ.vars.colors.text,
            i,
            HoW_QJ.vars.colors.section,
            LINK_STYLE_DEFAULT,
            zoneName,
            zoneName,
            HoW_QJ.vars.colors.text,
            HoW_QJ.vars.colors.complete,
            complete,
            HoW_QJ.vars.colors.text,
            HoW_QJ.vars.colors.value,
            stats.total,
            HoW_QJ.vars.colors.text
        ), 1, 1, 1, 1);
        i = i + 1
    end
    HoW_QJ:ScrollToFirstLine()
end

--- Update TextBuffer with quest info.
-- @param zoneName string the current selected zoneName.
-- @param questName string the current selected questName.
-- @return void
function HoW_QJ:PrintQuestInfo(zoneName, questName)
    local quest = HoW_QJ:GetQuestInfo(zoneName, questName)
    local questColor = HoW_QJ.vars.colors.unknown
    HoW_QJ_CT_TextBuffer:Clear()

    if quest then
        if quest.state == STATE.active then
            questColor = HoW_QJ.vars.colors.active
        elseif quest.state == STATE.complete then
            questColor = HoW_QJ.vars.colors.complete
        elseif quest.state == STATE.abandoned then
            questColor = HoW_QJ.vars.colors.abandoned
        end
    end
    HoW_QJ_CT_TextBuffer:AddMessage(string.format(
        "|c%s|H%d:quest:%s:%s:|h%s|h",
        questColor,
        LINK_STYLE_DEFAULT,
        zoneName, questName,
        questName
    ), 1, 1, 1, 1);
    if quest.level then
        HoW_QJ_CT_TextBuffer:AddMessage(string.format(
            "|c%sLevel: |c%s%d",
            HoW_QJ.vars.colors.text,
            HoW_QJ.vars.colors.value,
            quest.level
        ), 1, 1, 1, 1);
    end
    if quest.dungeon then
        HoW_QJ_CT_TextBuffer:AddMessage(string.format(
            "|c%sDungeon: |c%s%s",
            HoW_QJ.vars.colors.text,
            HoW_QJ.vars.colors.value,
            quest.dungeon
        ), 1, 1, 1, 1);
    end
    if quest.start then
        if quest.start.quest and quest.start.quest.description then
            HoW_QJ_CT_TextBuffer:AddMessage(string.format(
                "|c%s%s",
                HoW_QJ.vars.colors.text,
                quest.start.quest.description
            ), 1, 1, 1, 1);
        end
        if quest.start.agent then
            if quest.start.agent.name and quest.start.agent.mapname then
                HoW_QJ_CT_TextBuffer:AddMessage(string.format(
                    "|c%sQuest was given to me by |c%s%s |c%sin |c%s%s|c%s.",
                    HoW_QJ.vars.colors.text,
                    HoW_QJ.vars.colors.value,
                    quest.start.agent.name,
                    HoW_QJ.vars.colors.text,
                    HoW_QJ.vars.colors.value,
                    quest.start.agent.mapname,
                    HoW_QJ.vars.colors.text
                ), 1, 1, 1, 1);
            elseif quest.start.agent.mapname then
                HoW_QJ_CT_TextBuffer:AddMessage(string.format(
                    "|c%sQuest was given to me in |c%s%s|c%s.",
                    HoW_QJ.vars.colors.text,
                    HoW_QJ.vars.colors.value,
                    quest.start.agent.mapname,
                    HoW_QJ.vars.colors.text
                ), 1, 1, 1, 1);
            end
        end
    end

    if quest.finish then
        if quest.finish.agent and quest.finish.agent.name and quest.finish.agent.mapname then
            HoW_QJ_CT_TextBuffer:AddMessage(string.format(
                "|c%s%s |c%sin |c%s%s |c%sappreciated my efforts and rewarded me handsomely.",
                HoW_QJ.vars.colors.value,
                quest.finish.agent.name,
                HoW_QJ.vars.colors.text,
                HoW_QJ.vars.colors.value,
                quest.finish.agent.mapname,
                HoW_QJ.vars.colors.text
            ), 1, 1, 1, 1);
        end
        --if quest.finish.quest and quest.finish.quest.description then
        --    HoW_QJ_CT_TextBuffer:AddMessage(string.format(
        --        "|c%s%s",
        --        HoW_QJ.vars.colors.text,
        --        quest.finish.quest.description
        --    ), 1, 1, 1, 1);
        --end
    end
    -- TODO have to calculate a fair bit to get best fit..
    -- HoW_QJ:ScrollToLine(2)
end
--------------------------------------------------------------------------------
-- Update xml control functions.
--------------------------------------------------------------------------------

--- Set window anchor position.
-- @param anchor table with point, relativePoint, offsetX and offsetY.
-- @return void
function HoW_QJ:SetAnchor(anchor)
    HoW_QJ_CT:SetAnchor(anchor.point, GuiRoot, anchor.relativePoint, anchor.offsetX, anchor.offsetY)
end

--- Set window size.
-- @param dimension table with width and height
-- @return void
function HoW_QJ:SetDimensions(dimension)
    HoW_QJ_CT:SetDimensions(dimension.width, dimension.height)
end

--- Set window faded.
-- @param faded boolean faded state
-- @return void
function HoW_QJ:SetFaded(faded)
    HoW_QJ_CT_Backdrop:SetHidden(faded)
end


--- Fade window backdrop.
-- @return void
function HoW_QJ:FadeWindow()
    -- Remember window faded state.
    HoW_QJ.vars.faded = not HoW_QJ_CT_Backdrop:IsHidden()
    HoW_QJ:SetFaded(HoW_QJ.vars.faded)
end

--- Refresh window with current zone.
-- @return void
function HoW_QJ:RefreshWindow()
    HoW_QJ.Update()
    HoW_QJ.vars.currentZoneName = GetUnitZone("player")
    HoW_QJ:UpdateWindow(true)
end

--- Close the Quest Journal window.
-- @return void
function HoW_QJ:CloseWindow()
    -- Remember window state.
    HoW_QJ.vars.visible = false
    HoW_QJ_CT:SetHidden(true)
end

--- Hide the window if visible and screen state changed to some dialogue/menu interface.
-- @return void
function HoW_QJ:HideIfVisible()
    if HoW_QJ.vars.visible then
        HoW_QJ_CT:SetHidden(true)
    end
end

--- Show the window if visible and screen state returned to normal.
-- @return void
function HoW_QJ:ShowIfVisible()
    if HoW_QJ.vars.visible then
        HoW_QJ_CT:SetHidden(false)
    end
end

--------------------------------------------------------------------------------

-- Initialize addon.
EVENT_MANAGER:RegisterForEvent(HoW_QJ.name, EVENT_ADD_ON_LOADED, HoW_QJ.AddOnLoaded)
